/******************************************************************************
 *                      Sources BCB - www.developpez.com                      *
 * -------------------------------------------------------------------------- *
 *                                                                            *
 * LZWCDFunctions.cpp                                                         *
 * Version: 1.0.0 mise  jour le 13 fvrier 2005                              *
 * Copyright Cryptonyx (http://five.chez.tiscali.fr)                          *
 *                                                                            *
 * Test avec:                                                                *
 * SE: WinXP                                                                  *
 * IDE: BCB5                                                                  *
 *                                                                            *
 * -------------------------------------------------------------------------- *
 *                                                                            *
 * Cet utilitaire illustre une compression/dcompression avec l'algorithme    *
 * de Liv, Zempel et Welch (LZW)                                              *
 *                                                                            *
 * -------------------------------------------------------------------------- *
 *                                                                            *
 * Retrouvez d'autres ressources BCB sur :                                    *
 *  - http://c.developpez.com/sources/bcb/                                    *
 *  - http://c.developpez.com/faq/bcb/                                        *
 * Ainsi que sur les forums (http://www.developpez.net/forums/)               *
 *                                                                            *
 ******************************************************************************/

#include <SysUtils.hpp>
#include <stdio.h>
#include "LZWCDFunctions.h"

//---------------------------------------------------------------------------
#define LZW_INCORRECT_FILE_NAME             0
#define LZW_INPUT_FILE_OPEN_FAILURE         1
#define LZW_INPUT_FILE_READ_FAILURE         2
#define LZW_OUTPUT_FILE_OPEN_FAILURE        3
#define LZW_COMPRESSED                      4
#define LZW_DECOMPRESSION_FAILED            5
#define LZW_DECOMPRESSED                    6
#define LZW_NUMBER_OF_BYTES                 7
#define LZW_WRITE_FAILURE                   8
#define LZW_INCORRECT_HEADER_VERSION        9
#define LZW_INCORRECT_HEADER_STRUCTURE      10
#define LZW_ORIGINAL                        11
#define LZW_COMPRESSION_FAILED              12
#define LZW_UNDERCOMPRESSED                 13
#define LZW_CODE_GENERATION_FAILED          14
//---------------------------------------------------------------------------
// Choisir une langue avant compilation du code
//---------------------------------------------------------------------------
#define FRENCH // ou #define ENGLISH
#ifdef FRENCH
AnsiString ListOfErrorMessages[] = {
    "Nom de fichier incorrect",						// 00
    "Ouverture du fichier d'entre en lecture impossible",		// 01
    "Problme de lecture du fichier d'entre",				// 02
    "Ouverture du fichier de sortie en criture impossible",	        // 03
    " compress de ",							// 04
    "Erreur lors de la conversion des codes",   			// 05
    " dcompress en ",							// 06
    " octets",								// 07
    "Problme d'criture de ",						// 08
    "Version de l'entte incorrecte",					// 09
    "Structure de l'entte incorrecte",					// 10
    " % par rapport  l'original",					// 11
    "Fichier d'entre vide : pas de compression effectue", 		// 12
    " augment (entte volumineux) de ",                                // 13
    "Gnration des codes impossible",				        // 14
};
#endif
#ifdef ENGLISH
AnsiString ListOfErrorMessages[] = {
    "Incorrect file name",						// 00
    "Can't open input file in read mode",				// 01
    "Failure reading input file",					// 02
    "Can't open output file in write mode",				// 03
    " compressed at",							// 04
    "Fatal error during code expansion",				// 05
    " decompressed in ",						// 06
    " bytes",								// 07
    "Failure writing ",							// 08
    "Bad header version",						// 09
    "Bad header structure",						// 10
    " % with respect to the original file",				// 11
    "Input file empty: no compression performed", 			// 12
    " under compressed (large header) at ",				// 13
    "Failure generating codes",						// 14
};
#endif

//---------------------------------------------------------------------------
// Fonctions de la classe TLZWClass
//---------------------------------------------------------------------------
__fastcall TLZWClass::TLZWClass()
{
}

//---------------------------------------------------------------------------
//
// Initialize
//
//---------------------------------------------------------------------------
bool TLZWClass::Initialize(const AnsiString InputFile, const AnsiString OutputFile)
{
    if(InputFile == "" || OutputFile == "") {
    	ErrorCode   = eBadFileName;
        ErrorString = ListOfErrorMessages[LZW_INCORRECT_FILE_NAME];
        return false;
    }
    Source = InputFile;
    Destination = OutputFile;
    // Ouverture du fichier d'entre
    if((SourceFileHandle = FileOpen(InputFile, fmOpenRead)) <= 0) {
    	ErrorCode = eOpenReadFailed;
        ErrorString = ListOfErrorMessages[LZW_INPUT_FILE_OPEN_FAILURE];
        return false;
    }
    // Ouverture du fichier de sortie
    if((DestinationFileHandle = FileCreate(OutputFile)) < 0) {
        ErrorCode = eCreateFailed;
        ErrorString = ListOfErrorMessages[LZW_OUTPUT_FILE_OPEN_FAILURE];
        return false;
    }
    InputFileBuffer = (char *)NULL;
    OutputFileBuffer = (char *)NULL;
    CompressedSize = 0;
    Source = InputFile;
    Destination = OutputFile;
    NumberOfElements = 0;
    ExpandedSize = 0;
    EndOfHeader = 0;
    memset(CodeValueTable, 0, TABLE_SIZE);
    memset(PrefixCodeTable, 0, TABLE_SIZE);
    memset(AppendedCharacterTable, 0, TABLE_SIZE);
    memset(StringStack, 0, TABLE_SIZE);
    ErrorCode = eCorrect;
    ErrorString = "";
    return true;
}

//---------------------------------------------------------------------------
//
// Reset
//
//---------------------------------------------------------------------------
void TLZWClass::Reset()
{
    Source = "";
    Destination = "";
    if(SourceFileHandle)
        FileClose(SourceFileHandle);
    if(DestinationFileHandle)
        FileClose(DestinationFileHandle);
    SourceFileHandle = DestinationFileHandle = 0;
    if(InputFileBuffer)
        delete [] InputFileBuffer;
    if(OutputFileBuffer)
        delete [] OutputFileBuffer;
}

//---------------------------------------------------------------------------
//
// FileCompress
//
//	entre : nom du fichier d'entr, nom du fichier de sortie
//	sortie : vrai ou faux
//
//---------------------------------------------------------------------------
bool TLZWClass::FileCompress(const AnsiString InputFile, const AnsiString OutputFile)
{
unsigned long int i, j;
unsigned short int StringCode, index, NextCode = 256; // le premier code libre se trouve en 256

    if(Initialize(InputFile, OutputFile) == false || HeaderGenerate() == false)
        return false;

    if(HeaderGenerate() == false) {
        Reset();
        return false;
    }

    for(i = 0; i < TABLE_SIZE + 1; i++)
        CodeValueTable[i] = 0x8000;

    j = (unsigned long int)EndOfHeader;

    // Parcours du tampon d'entre caractre par caractre
    StringCode = InputFileBuffer[0];
    for(i = 1; i < ExpandedSize; i++) {
       // On vrifie la prsence de la chane dans la table.
       // Si elle y est, on copie sa valeur, sinon on l'ajoute.
       index = FindMatch(StringCode, InputFileBuffer[i]);
       if(CodeValueTable[index] != 0x8000)
          StringCode = CodeValueTable[index];
       else {
          if(NextCode <= MAX_CODE) {
             CodeValueTable[index] = NextCode++;
             PrefixCodeTable[index] = StringCode;
             AppendedCharacterTable[index] = InputFileBuffer[i];
          }
          // Quand une chane n'est pas dans la table, on la place dans le tampon
          j = OutputCode(j, StringCode);
          StringCode = InputFileBuffer[i];
       }
    }
    j = OutputCode(j, StringCode); // place le dernier code dans le tampon
    j = OutputCode(j, 0);       // vide le tampon
    if(FileWrite(DestinationFileHandle, OutputFileBuffer, j) != (int)j)
       ErrorCode = eWriteFailed;
    // la taille totale du tampon de sortie est j
    CompressedSize = j;
    ErrorCode = eCorrect;
    // Pourcentage de compression
    if((double)CompressedSize / ExpandedSize <= 1.0)
       ErrorString = Destination + ListOfErrorMessages[LZW_COMPRESSED]
       + IntToStr(100 - (100 * CompressedSize / ExpandedSize)) + ListOfErrorMessages[LZW_ORIGINAL];
    else
       ErrorString = Destination + ListOfErrorMessages[LZW_UNDERCOMPRESSED]
       + IntToStr((int)((100 * ((double)CompressedSize) / ExpandedSize) - 100)) + ListOfErrorMessages[LZW_ORIGINAL];

    Reset();
    if(ErrorCode == eWriteFailed)
        return false;
    return true;
}

//---------------------------------------------------------------------------
//
// FileDecompress
//
//	entre : noms des fichiers
//	sortie : vrai ou faux
//
//---------------------------------------------------------------------------
bool TLZWClass::FileDecompress(const AnsiString InputFile, const AnsiString OutputFile)
{
unsigned long int i, j = 0;
unsigned short int NewCode, OldCode, NextCode = 256; // le premier code libre est toujours 256
unsigned char CurrentCharacter;
int StackPosition;

    if(Initialize(InputFile, OutputFile) == false)
        return false;

    if(HeaderRead() == false) {
        Reset();
        return false;
    }
    // cration d'un tampon pour stocker les caractres dcods
    OutputFileBuffer = new unsigned char[ExpandedSize + 1];
    memset(OutputFileBuffer, 0, ExpandedSize);
    // lecture du premier code du fichier source
    i = (unsigned long int)EndOfHeader;
    OldCode  = InputCode(&i);
    CurrentCharacter = (unsigned char)OldCode;
    OutputFileBuffer[j++] = (unsigned char)OldCode;
    while(i < CompressedSize) {
        // Cette partie traite du cas particulier CHAINE + CARACTERE + CHAINE + CARACTERE + CHAINE
        // qui gnre un code indefini.  On dcode le dernier code et on ajoute un seul
        // caractre  la fin de la chane.
        NewCode  = InputCode(&i);
        if(NewCode >= NextCode) {
            StringStack[0] = CurrentCharacter;
            StackPosition = DecodeString(1, OldCode);
            if(StackPosition < 0)
                return false;
        }
        // Sinon on decode directement le nouveau code.
        else {
            StackPosition = DecodeString(0, NewCode);
            if(StackPosition < 0)
                return false;
        }
        // On crit la chane  l'envers dans le tampon de sortie en la dpilant
        CurrentCharacter = StringStack[StackPosition];
        while(StackPosition >= 0)
            OutputFileBuffer[j++] = StringStack[StackPosition--];
        // On ajoute un nouveau code si ncessaire dans la table des chanes
        if(NextCode <= MAX_CODE) {
            PrefixCodeTable[NextCode] = OldCode;
            AppendedCharacterTable[NextCode++] = CurrentCharacter;
        }
        OldCode = NewCode;
    }
    ErrorCode = eCorrect;
    ErrorString = Destination + ListOfErrorMessages[LZW_DECOMPRESSED]
       + IntToStr(ExpandedSize) + ListOfErrorMessages[LZW_NUMBER_OF_BYTES];

    OutputFileBuffer[ExpandedSize] = '\0';      // dimensionne le tampon  la taille relle du fichier original
    if(FileWrite(DestinationFileHandle, OutputFileBuffer, ExpandedSize) != (int)ExpandedSize) {
        ErrorString = ListOfErrorMessages[LZW_WRITE_FAILURE] + Destination;
        ErrorCode = eWriteFailed;
    }
    Reset();
    if(ErrorCode == eWriteFailed)
        return false;
    return true;
}

//---------------------------------------------------------------------------
//
//  GetLastError : renvoie le dernier message d'erreur
//
//---------------------------------------------------------------------------
AnsiString TLZWClass::GetLastError()
{
    return ErrorString;
}

//---------------------------------------------------------------------------
//
//  HeaderGenerate
//
//---------------------------------------------------------------------------
bool TLZWClass::HeaderGenerate()
{
LongOrChar FileSize;
ShortOrChar ShortCode;
AnsiString FileName;
unsigned long int i, j;
int DelimiterPosition;

    // Alloue un tampon de la taille du fichier et y charge son contenu
    ExpandedSize = FileSeek(SourceFileHandle, 0, 2);
    if(ExpandedSize == 0) {
        ErrorCode = eEmptyFile;
        ErrorString = ListOfErrorMessages[LZW_COMPRESSION_FAILED];
        return false;
    }
    FileSeek(SourceFileHandle, 0, 0);           // revient au dbut du fichier
    InputFileBuffer = new unsigned char[ExpandedSize + 1];
    memset(InputFileBuffer, 0, ExpandedSize);
    if((ExpandedSize = (unsigned long int)FileRead(SourceFileHandle, InputFileBuffer, ExpandedSize)) != ExpandedSize) {
        ErrorCode = eReadFailed;
        ErrorString = ListOfErrorMessages[LZW_INPUT_FILE_READ_FAILURE];
        return false;
    }
    // Allocation du tampon de sortie et mise  zro indispensable de tous ses bits
    // Le tampon de sortie doit tre plus grand que le tampon d'entre pour les fichiers
    // de petite taille  cause de l'adjonction des paramtres lis au fichier d'entre
    j= (unsigned long int)(1.5 * ExpandedSize);
    if(ExpandedSize < 1024) {
        OutputFileBuffer = new unsigned char[301 + j];
        memset(OutputFileBuffer, 0, 300 + j);
    }
    else {
        // dans certains cas (ex. les .pdf) le fichier de sortie est plus gros
        OutputFileBuffer = new unsigned char[j];
        memset(OutputFileBuffer, 0, j - 1);
    }

    // On saute 2 octets pour sauvegarder la taille de l'entte
    // Ecriture du numro de version
    ShortCode.ShortValue = VERSION;
    OutputFileBuffer[2] = ShortCode.CharSize[0];
    OutputFileBuffer[3] = ShortCode.CharSize[1];

    // Ecriture de la taille du fichier en octets
    FileSize.LongSize = ExpandedSize;
    for(i = 4; i < 8; i++)
        OutputFileBuffer[i] = FileSize.CharSize[i - 4];

    // Ecriture du nom du fichier original sans le chemin
    DelimiterPosition = Source.LastDelimiter(PATH_DELIMITER);
    if(DelimiterPosition < 1)
        // Il n'y a pas le chemin
        FileName = Source;
    else
        FileName = Source.SubString(DelimiterPosition + 1, Source.Length() - DelimiterPosition);

    // Ecriture de la taille du nom du fichier original - on coupe  255 caractres
    if((DelimiterPosition = FileName.Length()) > 255)
        DelimiterPosition = 255;
    OutputFileBuffer[i++] = (unsigned char)DelimiterPosition;
    for(j = 1; j <= (unsigned long int)DelimiterPosition; j++)
        OutputFileBuffer[i++] = FileName[j];

    EndOfHeader = (unsigned short int)i;
    // Ecriture de la taille de l'entte dans les 2 premiers octets du tampon
    ShortCode.ShortValue = EndOfHeader;
    OutputFileBuffer[0] = ShortCode.CharSize[0];
    OutputFileBuffer[1] = ShortCode.CharSize[1];
    ErrorCode = eCorrect;
    return true;
}

//---------------------------------------------------------------------------
//
//	HeaderRead
//
//---------------------------------------------------------------------------
bool TLZWClass::HeaderRead()
{
LongOrChar FileSize;
ShortOrChar ShortCode;
int i;
unsigned short int j;
char pszShortBuffer[256];

    CompressedSize = FileSeek(SourceFileHandle, 0, 2);
    FileSeek(SourceFileHandle, 0, 0);
    InputFileBuffer = new unsigned char[CompressedSize + 1];
    if(FileRead(SourceFileHandle, InputFileBuffer, CompressedSize) != (int)CompressedSize) {
        ErrorCode   = eOpenReadFailed;
        ErrorString = ListOfErrorMessages[LZW_INPUT_FILE_READ_FAILURE];
        return false;
    }

    // Lecture de la taille de l'entte pour contrle
    ShortCode.CharSize[0] = InputFileBuffer[0];
    ShortCode.CharSize[1] = InputFileBuffer[1];
    EndOfHeader = ShortCode.ShortValue;

    // Lecture du numro de version
    ShortCode.CharSize[0] = InputFileBuffer[2];
    ShortCode.CharSize[1] = InputFileBuffer[3];
    if(ShortCode.ShortValue < VERSION) {
        // Mauvaise version
        ErrorCode   = eBadVersion;
        ErrorString = ListOfErrorMessages[LZW_INCORRECT_HEADER_VERSION];
        return false;
    }

    // Lecture de la taille du fichier en octets
    for(i = 4; i < 8; i++)
        FileSize.CharSize[i - 4] = InputFileBuffer[i];
    ExpandedSize = FileSize.LongSize;

    // Lecture de la taille du nom du fichier original
    ShortCode.ShortValue = (unsigned short int)InputFileBuffer[i++];

    for(j = 0; j < ShortCode.ShortValue; j++)
        pszShortBuffer[j] = InputFileBuffer[i++];

    // Zero terminal
    pszShortBuffer[ShortCode.ShortValue] = 0;
    if(Destination == "")
        Destination.sprintf("%s", pszShortBuffer);

    EndOfHeader = (unsigned short int)i;
    ErrorCode = eCorrect;
    ErrorString = "";
    return true;
}

//---------------------------------------------------------------------------
// FindMatch
//      Fonction de hachage qui permet de trouver une entre dans la table
//      pour une chane de caractres. Si une entre est trouve, la fonction
//      renvoie sa position. Si la chane n'est pas trouve, la position de la
//      premire entre libre est renvoye.
//---------------------------------------------------------------------------
unsigned short int TLZWClass::FindMatch(unsigned short int HashPrefix, unsigned short int HashCharacter)
{
int index;
int offset;

  index = (HashCharacter << HASHING_SHIFT) ^ HashPrefix;
  if(index == 0)
    offset = 1;
  else
    offset = TABLE_SIZE - index;

  while(true) {
    if(CodeValueTable[index] == 0x8000)
      return((unsigned short int)index);

    if((PrefixCodeTable[index] == HashPrefix) && (AppendedCharacterTable[index] == (unsigned char)HashCharacter))
      return((unsigned short int)index);

    index -= offset;
    if(index < 0)
      index += TABLE_SIZE;
  }
}

//---------------------------------------------------------------------------
// InputCode
//      Lit des codes de longueur BITS depuis le tampon d'entre
//---------------------------------------------------------------------------
unsigned short int TLZWClass::InputCode(unsigned long int *CodePosition)
{
unsigned short int ReturnCode;                  // un code fait entre 12 et 14 bits donc il tient sur 2 octets
static unsigned short int InputBitCount = 0;
static unsigned short int InputBitBuffer = 0;

  while(InputBitCount <= 8) {
    InputBitBuffer |= (unsigned short int)InputFileBuffer[(*CodePosition)++] << (8 - InputBitCount);
    InputBitCount += 8;
  }
  ReturnCode = InputBitBuffer >> (16 - BITS);
  InputBitBuffer <<= BITS;
  InputBitCount -= BITS;
  return(ReturnCode);
}

//---------------------------------------------------------------------------
// OutputCode
//      Ecrit des codes de longueur variable dans le tampon de sortie
//---------------------------------------------------------------------------
unsigned long int TLZWClass::OutputCode(unsigned long int CodePosition, unsigned short int Code)
{
static unsigned short int OutputBitCount = 0;
static unsigned short int OutputBitBuffer = 0;

  OutputBitBuffer |= (unsigned short int)Code << (16 - BITS - OutputBitCount);
  OutputBitCount += BITS;
  while(OutputBitCount >= 8) {
    // en toute rigueur, il faudrait tester CodePosition
    // qui ne doit pas excder la taille du tampon de sortie
    // en paratique, le tampon de sortie a t surdimensionn
    OutputFileBuffer[CodePosition++] = (unsigned char)(OutputBitBuffer >> 8);
    OutputBitBuffer <<= 8;
    OutputBitCount -= 8;
  }
  return((unsigned long int)CodePosition);
}

//---------------------------------------------------------------------------
// DecodeString
//      Decode une chane de la table et la place dans une pile.
//      La pile peut alors tre lue dans l'ordre inverse lors de la dcompression.
//---------------------------------------------------------------------------
int TLZWClass::DecodeString(int StackDepth, unsigned int Code)
{
   // prconditions
   if((StackDepth < 0) || (StackDepth >= MAX_VALUE)) {
      ErrorCode = eBadCode;
      ErrorString = ListOfErrorMessages[LZW_DECOMPRESSION_FAILED];
      return(-1);
   }
   while(Code > 255) {
        StringStack[StackDepth++] = AppendedCharacterTable[Code];
        Code = PrefixCodeTable[Code];
        if(StackDepth >= MAX_VALUE) {
           ErrorCode = eBadCode;
           ErrorString = ListOfErrorMessages[LZW_DECOMPRESSION_FAILED];
           return(-1);
        }
  }
  StringStack[StackDepth] = (unsigned char)Code;
  return(StackDepth);
}

